"""OpenAPI spec renderer."""

import collections.abc
import copy
import functools
import http.client
import json

import deepmerge
import docutils.parsers.rst.directives as directives
import m2r
import requests
import sphinx.util.logging as logging

from sphinxcontrib.openapi import _lib2to3 as lib2to3
from sphinxcontrib.openapi.renderers import abc
from sphinxcontrib.openapi.schema_utils import example_from_schema


CaseInsensitiveDict = requests.structures.CaseInsensitiveDict


logger = logging.getLogger(__name__)


def indented(generator, indent=3):
    for item in generator:
        if item:
            item = " " * indent + item
        yield item


def _iterinorder(iterable, order_by, key=lambda x: x, case_sensitive=False):
    """Iterate over iterable in a given order."""

    order_by = collections.defaultdict(
        # Assume default priority is `Infinity` which means the lowest one.
        # This value is effectively used if there's no corresponding value in a
        # given 'order_by' array.
        lambda: float("Inf"),
        # Passed 'order_by' may be 'None' which means *do not reorder, use
        # natural order*. In order to avoid special cases in the code, we're
        # simply falling back to an empty 'order_by' array since it effectively
        # means *assume every item in 'iterable' has equal priority*.
        ((value, i) for i, value in enumerate(order_by or [])),
    )
    yield from sorted(
        iterable,
        key=lambda value: order_by[
            key(value) if case_sensitive else key(value).lower()
        ],
    )


def _iterexamples(media_types, example_preference, examples_from_schemas):
    """Iterate over examples and return them according to the caller preference."""

    for content_type in _iterinorder(media_types, example_preference):
        media_type = media_types[content_type]

        # Look for a example in a bunch of possible places. According to
        # OpenAPI v3 spec, `examples` and `example` keys are mutually
        # exclusive, so there's no much difference between their
        # inspection order, while both must take precedence over a
        # schema example.
        if media_type.get("examples", {}):
            for example in media_type["examples"].values():
                if "externalValue" in example:
                    if not example["externalValue"].startswith(("http://", "https://")):
                        logger.warning(
                            "Not supported protocol in 'externalValue': %s",
                            example["externalValue"],
                        )
                        continue

                    try:
                        response = requests.get(example["externalValue"])
                        response.raise_for_status()

                        example["value"] = response.text
                        example.pop("externalValue")
                    except Exception:
                        logger.error(
                            "Cannot retrieve example from: '%s'",
                            example["externalValue"],
                        )
                        continue
                break
            else:
                # If the loop over examples has not been interrupted, we
                # probably didn't find an example to render. In that case,
                # let's try and go next media type.
                continue
        elif media_type.get("example"):
            # Save example from "example" in "examples" compatible format. This
            # allows to treat all returned examples the same way.
            example = {"value": media_type["example"]}
        elif media_type.get("schema", {}).get("example"):
            # Save example from "schema" in "examples" compatible format. This
            # allows to treat all returned examples the same way.
            example = {"value": media_type["schema"]["example"]}
        elif "schema" in media_type and examples_from_schemas:
            # Convert schema to example
            example = {"value": example_from_schema(media_type["schema"])}
            pass
        else:
            continue

        yield content_type, example


def _get_markers_from_object(oas_object, schema):
    """Retrieve a bunch of OAS object markers."""

    markers = []

    schema_type = _get_schema_type(schema)
    if schema_type:
        if schema.get("format"):
            schema_type = f"{schema_type}:{schema['format']}"
        elif schema.get("enum"):
            schema_type = f"{schema_type}:enum"
        markers.append(schema_type)
    elif schema.get("enum"):
        markers.append("enum")

    if oas_object.get("required"):
        markers.append("required")

    if oas_object.get("deprecated"):
        markers.append("deprecated")

    if schema.get("deprecated"):
        markers.append("deprecated")

    return markers


def _is_json_mimetype(mimetype):
    """Returns 'True' if a given mimetype implies JSON data."""

    return any(
        [
            mimetype == "application/json",
            mimetype.startswith("application/") and mimetype.endswith("+json"),
        ]
    )


def _is_2xx_status(status_code):
    """Returns 'True' if a given status code is one of successful."""

    return str(status_code).startswith("2")


def _get_schema_type(schema):
    """Retrieve schema type either by reading 'type' or guessing."""

    # There are a lot of OpenAPI specs out there that may lack 'type' property
    # in their schemas. I fount no explanations on what is expected behaviour
    # in this case neither in OpenAPI nor in JSON Schema specifications. Thus
    # let's assume what everyone assumes, and try to guess schema type at least
    # for two most popular types: 'object' and 'array'.
    if "type" not in schema:
        if "properties" in schema:
            schema_type = "object"
        elif "items" in schema:
            schema_type = "array"
        else:
            schema_type = None
    else:
        schema_type = schema["type"]
    return schema_type


_merge_mappings = deepmerge.Merger(
    [(collections.abc.Mapping, deepmerge.strategy.dict.DictStrategies("merge"))],
    ["override"],
    ["override"],
).merge


class HttpdomainRenderer(abc.RestructuredTextRenderer):
    """Render OpenAPI v3 using `sphinxcontrib-httpdomain` extension."""

    _markup_converters = {"commonmark": m2r.convert, "restructuredtext": lambda x: x}
    _response_examples_for = {"200", "201", "202", "2XX"}
    _request_parameters_order = ["header", "path", "query", "cookie"]

    option_spec = {
        "markup": functools.partial(directives.choice, values=_markup_converters),
        "http-methods-order": lambda s: s.split(),
        "response-examples-for": None,
        "request-parameters-order": None,
        "example-preference": None,
        "request-example-preference": None,
        "response-example-preference": None,
        "generate-examples-from-schemas": directives.flag,
        "no-json-schema-description": directives.flag,
    }

    def __init__(self, state, options):
        super().__init__(state, options)

        self._convert_markup = self._markup_converters[
            options.get("markup", "commonmark")
        ]
        self._http_methods_order = [
            http_method.lower() for http_method in options.get("http-methods-order", [])
        ]
        self._response_examples_for = options.get(
            "response-examples-for", self._response_examples_for
        )
        self._request_parameters_order = [
            parameter_type.lower()
            for parameter_type in options.get(
                "request-parameters-order", self._request_parameters_order
            )
        ]
        self._example_preference = options.get("example-preference")
        self._request_example_preference = options.get(
            "request-example-preference", self._example_preference
        )
        self._response_example_preference = options.get(
            "response-example-preference", self._example_preference
        )
        self._generate_example_from_schema = "generate-examples-from-schemas" in options
        self._json_schema_description = "no-json-schema-description" not in options

    def render_restructuredtext_markup(self, spec):
        """Spec render entry point."""

        if spec.get("swagger") == "2.0":
            spec = lib2to3.convert(spec)
        yield from self.render_paths(spec.get("paths", {}))

    def render_paths(self, paths):
        """Render OAS paths item."""

        for endpoint, path in paths.items():
            common_parameters = path.pop("parameters", [])

            # OpenAPI's path description may contain objects of different
            # types. Since we're interested in rendering only objects of
            # operation type, let's remove irrelevant one from the definition
            # in order to simplify further code.
            for key in {"summary", "description", "servers"}:
                path.pop(key, None)

            for method in _iterinorder(path, self._http_methods_order):
                operation = path[method]
                operation.setdefault("parameters", [])
                operation_parameters_ids = set(
                    (parameter["name"], parameter["in"])
                    for parameter in operation["parameters"]
                )
                operation["parameters"] = [
                    parameter
                    for parameter in common_parameters
                    if (parameter["name"], parameter["in"])
                    not in operation_parameters_ids
                ] + operation["parameters"]

                yield from self.render_operation(endpoint, method, operation)
                yield ""

    def render_operation(self, endpoint, method, operation):
        """Render OAS operation item."""

        yield f".. http:{method}:: {endpoint}"

        if operation.get("deprecated"):
            yield f"   :deprecated:"
        yield f""

        if operation.get("summary"):
            yield f"   **{operation['summary']}**"
            yield f""

        if operation.get("description"):
            yield from indented(
                self._convert_markup(operation["description"]).strip().splitlines()
            )
            yield f""

        yield from indented(self.render_parameters(operation.get("parameters", [])))
        if "requestBody" in operation:
            yield from indented(
                self.render_request_body(operation["requestBody"], endpoint, method)
            )
        yield from indented(self.render_responses(operation["responses"]))

    def render_parameters(self, parameters):
        """Render OAS operation's parameters."""

        for parameter in _iterinorder(
            parameters, self._request_parameters_order, key=lambda value: value["in"]
        ):
            yield from self.render_parameter(parameter)

    def render_parameter(self, parameter):
        """Render OAS operation's parameter."""

        kinds = CaseInsensitiveDict(
            {"path": "param", "query": "queryparam", "header": "reqheader"}
        )
        schema = parameter.get("schema", {})

        if "content" in parameter:
            # According to OpenAPI v3 spec, 'content' in this case may
            # have one and only one entry. Hence casting its values to
            # list is not expensive and should be acceptable.
            schema = list(parameter["content"].values())[0].get("schema", {})

        if parameter["in"] not in kinds:
            logger.warning(
                "OpenAPI spec contains parameter '%s' (in: '%s') that cannot "
                "be rendererd.",
                parameter["name"],
                parameter["in"],
            )
            return

        yield f":{kinds[parameter['in']]} {parameter['name']}:"

        if parameter.get("description"):
            yield from indented(
                self._convert_markup(parameter["description"]).strip().splitlines()
            )

        markers = _get_markers_from_object(parameter, schema)
        if markers:
            markers = ", ".join(markers)
            yield f":{kinds[parameter['in']]}type {parameter['name']}: {markers}"

    def render_request_body(self, request_body, endpoint, method):
        """Render OAS operation's requestBody."""

        if self._json_schema_description:
            for content_type, content in request_body["content"].items():
                if _is_json_mimetype(content_type) and content.get("schema"):
                    yield from self.render_json_schema_description(
                        content["schema"], "req"
                    )
                    yield ""
                    break

        yield from self.render_request_body_example(request_body, endpoint, method)
        yield ""

    def render_request_body_example(self, request_body, endpoint, method):
        """Render OAS operation's requestBody's example."""

        content_type, example = next(
            _iterexamples(
                request_body["content"],
                self._request_example_preference,
                self._generate_example_from_schema,
            ),
            (None, None),
        )

        if content_type and example:
            example = example["value"]

            if not isinstance(example, str):
                example = json.dumps(example, indent=2)

            yield f".. sourcecode:: http"
            yield f""
            yield f"   {method.upper()} {endpoint} HTTP/1.1"
            yield f"   Content-Type: {content_type}"
            yield f""
            yield from indented(example.splitlines())

    def render_responses(self, responses):
        """Render OAS operation's responses."""

        if self._json_schema_description:
            for status_code, response in responses.items():
                if _is_2xx_status(status_code):
                    for content_type, content in response.get("content", {}).items():
                        if _is_json_mimetype(content_type) and content.get("schema"):
                            yield from self.render_json_schema_description(
                                content["schema"], "res"
                            )
                            yield ""
                            break
                    break

        for status_code, response in responses.items():
            # Due to the way how YAML spec is parsed, status code may be
            # infered as integer. In order to spare some cycles on type
            # guessing going on, let's ensure it's always string at this point.
            yield from self.render_response(str(status_code), response)

    def render_response(self, status_code, response):
        """Render OAS operation's response."""

        yield f":statuscode {status_code}:"
        yield from indented(
            self._convert_markup(response["description"]).strip().splitlines()
        )

        if "content" in response and status_code in self._response_examples_for:
            yield ""
            yield from indented(
                self.render_response_example(response["content"], status_code)
            )

        if "headers" in response:
            yield ""

            for header_name, header_value in response["headers"].items():
                # According to OpenAPI v3 specification, if a response header
                # is defined with the name 'Content-Type', it shall be ignored.
                if header_name.lower() == "content-type":
                    continue

                yield f":resheader {header_name}:"

                if header_value.get("description"):
                    yield from indented(
                        self._convert_markup(header_value["description"])
                        .strip()
                        .splitlines()
                    )

                schema = header_value.get("schema", {})
                if "content" in header_value:
                    # According to OpenAPI v3 spec, 'content' in this case may
                    # have one and only one entry. Hence casting its values to
                    # list is not expensive and should be acceptable.
                    schema = list(header_value["content"].values())[0].get("schema", {})

                markers = _get_markers_from_object(header_value, schema)
                if markers:
                    markers = ", ".join(markers)
                    yield f":resheadertype {header_name}: {markers}"

    def render_response_example(self, media_type, status_code):
        # OpenAPI 3.0 spec may contain more than one response media type, and
        # each media type may contain more than one example. Rendering all
        # invariants normally is not an option because the result will be hard
        # to read and follow. The best option we can go with at this moment is
        # to render first found example of either response media type. Users
        # should control what to render by putting recommended example first in
        # the list.
        content_type, example = next(
            _iterexamples(
                media_type,
                self._response_example_preference,
                self._generate_example_from_schema,
            ),
            (None, None),
        )

        if content_type and example:
            example = example["value"]

            if not isinstance(example, str):
                example = json.dumps(example, indent=2)

            # According to OpenAPI v3 spec, status code may be a special value
            # - "default". It's not quite clear what to render in this case.
            # One possible option is to avoid rendering status code at all.
            # This option, however, suffers from broken code highlighting
            # because Pygments relies on the snippet to start with HTTP
            # protocol line. That said, probably the best we can do at the
            # moment is to render some generic status.
            if status_code == "default":
                status_code = "000"
                status_text = "Reason-Phrase"
            else:
                # According to OpenAPI v3 spec, status code may define a range
                # of response codes. Since we're talking about rendered example
                # here, we may show either code from range, but for the sake of
                # simplicity let's pick the first one.
                status_code = status_code.replace("XX", "00")
                status_text = http.client.responses.get(int(status_code), "-")

            yield f".. sourcecode:: http"
            yield f""
            yield f"   HTTP/1.1 {status_code} {status_text}"
            yield f"   Content-Type: {content_type}"
            yield f""
            yield from indented(example.splitlines())

    def render_json_schema_description(self, schema, req_or_res):
        """Render JSON schema's description."""

        def _resolve_combining_schema(schema):
            if "oneOf" in schema:
                # The part with merging is a vague one since I only found a
                # single 'oneOf' example where such merging was assumed, and no
                # explanations in the spec itself.
                merged_schema = schema.copy()
                merged_schema.update(merged_schema.pop("oneOf")[0])
                return merged_schema

            elif "anyOf" in schema:
                # The part with merging is a vague one since I only found a
                # single 'oneOf' example where such merging was assumed, and no
                # explanations in the spec itself.
                merged_schema = schema.copy()
                merged_schema.update(merged_schema.pop("anyOf")[0])
                return merged_schema

            elif "allOf" in schema:
                # Since the item is represented by all schemas from the array,
                # the best we can do is to render them all at once
                # sequentially. Please note, the only way the end result will
                # ever make sense is when all schemas from the array are of
                # object type.
                merged_schema = schema.copy()
                for item in merged_schema.pop("allOf"):
                    merged_schema = _merge_mappings(merged_schema, copy.deepcopy(item))
                return merged_schema

            elif "not" in schema:
                # Eh.. do nothing because I have no idea what can we do.
                return {}

            return schema

        def _traverse_schema(schema, name, is_required=False):
            schema_type = _get_schema_type(schema)

            if {"oneOf", "anyOf", "allOf"} & schema.keys():
                # Since an item can represented by either or any schema from
                # the array of schema in case of `oneOf` and `anyOf`
                # respectively, the best we can do for them is to render the
                # first found variant. In other words, we are going to traverse
                # only a single schema variant and leave the rest out. This is
                # by design and it was decided so in order to keep produced
                # description clear and simple.
                yield from _traverse_schema(_resolve_combining_schema(schema), name)

            elif "not" in schema:
                yield name, {}, is_required

            elif schema_type == "object":
                if name:
                    yield name, schema, is_required

                required = set(schema.get("required", []))

                for key, value in schema.get("properties", {}).items():
                    # In case of the first recursion call, when 'name' is an
                    # empty string, we should go with 'key' only in order to
                    # avoid leading dot at the beginning.
                    yield from _traverse_schema(
                        value,
                        f"{name}.{key}" if name else key,
                        is_required=key in required,
                    )

            elif schema_type == "array":
                yield from _traverse_schema(schema["items"], f"{name}[]")

            elif "enum" in schema:
                yield name, schema, is_required

            elif schema_type is not None:
                yield name, schema, is_required

        schema = _resolve_combining_schema(schema)
        schema_type = _get_schema_type(schema)

        # On root level, httpdomain supports only 'object' and 'array' response
        # types. If it's something else, let's do not even try to render it.
        if schema_type not in {"object", "array"}:
            return

        # According to httpdomain's documentation, 'reqjsonobj' is an alias for
        # 'reqjson'. However, since the same name is passed as a type directive
        # internally, it actually can be used to specify its type. The same
        # goes for 'resjsonobj'.
        directives_map = {
            "req": {
                "object": ("reqjson", "reqjsonobj"),
                "array": ("reqjsonarr", "reqjsonarrtype"),
            },
            "res": {
                "object": ("resjson", "resjsonobj"),
                "array": ("resjsonarr", "resjsonarrtype"),
            },
        }

        # These httpdomain's fields always expect either JSON Object or JSON
        # Array. No primitive types are allowed as input.
        directive, typedirective = directives_map[req_or_res][schema_type]

        # Since we use JSON array specific httpdomain directives if a schema
        # we're about to render is an array, there's no need to render that
        # array in the first place.
        if schema_type == "array":
            schema = schema["items"]

            # Even if a root element is an array, items it contain must not be
            # of a primitive types.
            if _get_schema_type(schema) not in {"object", "array"}:
                return

        for name, schema, is_required in _traverse_schema(schema, ""):
            yield f":{directive} {name}:"

            if schema.get("description"):
                yield from indented(
                    self._convert_markup(schema["description"]).strip().splitlines()
                )

            markers = _get_markers_from_object({}, schema)

            if is_required:
                markers.append("required")

            if markers:
                markers = ", ".join(markers)
                yield f":{typedirective} {name}: {markers}"



""".convert_operation() test suite."""

import pytest

import sphinxcontrib.openapi._lib2to3 as lib2to3


@pytest.fixture(scope="function")
def convert_operation(oas_fragment):
    def _wrapper(operation):
        oas2 = oas_fragment(
            """
            swagger: "2.0"
            info:
              title: An example spec
              version: "1.0"
            paths:
              /test:
                get:
                  responses:
                    '200':
                      description: a response description
            """
        )
        oas2["paths"]["/test"]["get"] = operation

        oas3 = lib2to3.convert(oas2)
        return oas3["paths"]["/test"]["get"]

    return _wrapper


def test_minimal(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            responses:
              '200':
                description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        responses:
          '200':
            description: a response description
        """
    )


def test_complete(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            tags:
              - tag_a
              - tag_b
            summary: an operation summary
            description: an operation description
            externalDocs: https://docs.example.com/
            operationId: myOperation
            produces:
              - application/json
            parameters:
              - in: header
                name: token
                type: string
              - in: path
                name: username
                required: true
                type: string
              - in: query
                name: id
                type: string
            deprecated: false
            responses:
              '200':
                description: a response description
                schema:
                  items:
                    format: int32
                    type: integer
                  type: array
            """
        ),
    )
    assert converted == oas_fragment(
        """
        tags:
          - tag_a
          - tag_b
        summary: an operation summary
        description: an operation description
        externalDocs: https://docs.example.com/
        operationId: myOperation
        parameters:
          - in: header
            name: token
            schema:
              type: string
          - in: path
            name: username
            required: true
            schema:
              type: string
          - in: query
            name: id
            schema:
              type: string
        deprecated: false
        responses:
          '200':
            content:
              application/json:
                schema:
                  items:
                    format: int32
                    type: integer
                  type: array
            description: a response description
        """
    )


def test_request_body(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            description: an operation description
            consumes:
              - application/json
            parameters:
              - in: path
                name: username
                required: true
                type: string
              - in: body
                name: inventory
                schema:
                  type: object
                  properties:
                    t-shirt:
                      type: boolean
                required: true
            responses:
              '200':
                description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        description: an operation description
        parameters:
          - in: path
            name: username
            required: true
            schema:
              type: string
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  t-shirt:
                    type: boolean
          required: true
        responses:
          '200':
             description: a response description
        """
    )


def test_request_body_formdata(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            description: an operation description
            consumes:
              - application/x-www-form-urlencoded
            parameters:
              - in: path
                name: username
                required: true
                type: string
              - in: formData
                name: t-shirt
                type: boolean
            responses:
              '200':
                description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        description: an operation description
        parameters:
          - in: path
            name: username
            required: true
            schema:
              type: string
        requestBody:
          content:
            application/x-www-form-urlencoded:
              schema:
                type: object
                properties:
                  t-shirt:
                    type: boolean
        responses:
          '200':
             description: a response description
        """
    )


def test_only_request_body(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            description: an operation description
            consumes:
              - application/json
            parameters:
              - in: body
                name: inventory
                schema:
                  type: object
                  properties:
                    t-shirt:
                      type: boolean
                required: true
            responses:
              '200':
                description: a response description
            """
        ),
    )
    assert converted == oas_fragment(
        """
        description: an operation description
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  t-shirt:
                    type: boolean
          required: true
        responses:
          '200':
             description: a response description
        """
    )


def test_vendor_extensions(convert_operation, oas_fragment):
    converted = convert_operation(
        oas_fragment(
            """
            responses:
              '200':
                description: a response description
            x-vendor-ext: vendor-ext
            """
        ),
    )
    assert converted == oas_fragment(
        """
        responses:
          '200':
            description: a response description
        x-vendor-ext: vendor-ext
        """
    )



"""OpenAPI spec renderer: render_response_example."""

import textwrap

import pytest
import responses

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


@pytest.mark.parametrize(
    ["media_type"],
    [
        pytest.param(
            """
            application/json:
              examples:
                test:
                  value:
                    foo: bar
                    baz: 42
            """,
            id="examples",
        ),
        pytest.param(
            """
            application/json:
              example:
                foo: bar
                baz: 42
            """,
            id="example",
        ),
        pytest.param(
            """
            application/json:
              schema:
                example:
                  foo: bar
                  baz: 42
            """,
            id="schema/example",
        ),
        pytest.param(
            """
            application/json:
              examples:
                test:
                  value: |
                    {
                      "foo": "bar",
                      "baz": 42
                    }
            """,
            id="examples::str",
        ),
        pytest.param(
            """
            application/json:
              example: |
                {
                  "foo": "bar",
                  "baz": 42
                }
            """,
            id="example::str",
        ),
        pytest.param(
            """
            application/json:
              schema:
                example: |
                  {
                    "foo": "bar",
                    "baz": 42
                  }
            """,
            id="schema/example::str",
        ),
        pytest.param(
            """
            application/json:
              schema:
                example:
                  foobar: bazinga
              example:
                foo: bar
                baz: 42
            """,
            id="example-beats-schema/example",
        ),
        pytest.param(
            """
            application/json:
              schema:
                example:
                  foobar: bazinga
              examples:
                test:
                  value:
                    foo: bar
                    baz: 42
            """,
            id="examples-beats-schema/example",
        ),
    ],
)
def test_render_response_example(testrenderer, oas_fragment, media_type):
    """Path response's example is rendered."""

    markup = textify(
        testrenderer.render_response_example(oas_fragment(media_type), "200")
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "foo": "bar",
             "baz": 42
           }
        """.rstrip()
    )


def test_render_response_example_1st_from_examples(testrenderer, oas_fragment):
    """Path response's first example is rendered."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  examples:
                    foo:
                      value:
                        foo: bar
                        baz: 42
                    bar:
                      value:
                        foobar: bazinga
                """
            ),
            "200",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "foo": "bar",
             "baz": 42
           }
        """.rstrip()
    )


def test_render_response_example_1st_from_media_type(testrenderer, oas_fragment):
    """Path response's example from first media type is rendered."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/plain:
                  example: |
                    foo = "bar"
                    baz = 42
                application/json:
                  schema:
                    type: object
                """
            ),
            "200",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["example_preference_key"],
    [pytest.param("response-example-preference"), pytest.param("example-preference")],
)
def test_render_response_example_preference(
    fakestate, example_preference_key, oas_fragment
):
    """Path response's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {example_preference_key: ["text/plain"]}
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  example:
                    foo: bar
                    baz: 42
                text/plain:
                  example: |
                    foo = "bar"
                    baz = 42
                """
            ),
            "200",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["example_preference_key"],
    [pytest.param("response-example-preference"), pytest.param("example-preference")],
)
def test_render_response_example_preference_complex(
    fakestate, example_preference_key, oas_fragment
):
    """Path response's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {example_preference_key: ["application/json", "text/plain"]}
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                text/plain:
                  example: |
                    foo = "bar"
                    baz = 42
                application/json:
                  schema:
                    type: object
                """
            ),
            "200",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


def test_render_response_example_preference_priority(fakestate, oas_fragment):
    """Path response's example from preferred media type is rendered."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate,
        {
            "example-preference": ["application/json"],
            "response-example-preference": ["text/plain"],
        },
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  example:
                    foo: bar
                    baz: 42
                text/plain:
                  example: |
                    foo = "bar"
                    baz = 42
                """
            ),
            "200",
        )
    )

    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/plain

           foo = "bar"
           baz = 42
        """.rstrip()
    )


@responses.activate
def test_render_response_example_external(testrenderer, oas_fragment):
    """Path response's example can be retrieved from external location."""

    responses.add(
        responses.GET,
        "https://example.com/json/examples/test.json",
        json={"foo": "bar", "baz": 42},
        status=200,
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  examples:
                    test:
                      externalValue: https://example.com/json/examples/test.json
                """
            ),
            "200",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {"foo": "bar", "baz": 42}
        """.rstrip()
    )


@responses.activate
def test_render_response_example_external_errored_next_example(
    testrenderer, caplog, oas_fragment
):
    """Path response's example fallbacks on next when external cannot be retrieved."""

    responses.add(
        responses.GET, "https://example.com/json/examples/test.json", status=404,
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  examples:
                    test:
                      externalValue: https://example.com/json/examples/test.json
                    fallback:
                      value: '{"spam": 42}'
                """
            ),
            "200",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {"spam": 42}
        """.rstrip()
    )


@responses.activate
def test_render_response_example_external_errored_next_media_type(
    testrenderer, oas_fragment, caplog
):
    """Path response's example fallbacks on next when external cannot be retrieved."""

    responses.add(
        responses.GET, "https://example.com/json/examples/test.json", status=404,
    )

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  examples:
                    test:
                      externalValue: https://example.com/json/examples/test.json
                text/csv:
                  example: spam,42
                """
            ),
            "200",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/csv

           spam,42
        """.rstrip()
    )


def test_render_response_example_content_type(testrenderer, oas_fragment):
    """Path response's example can render something other than application/json."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                """
            ),
            "200",
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


def test_render_response_example_noop(testrenderer, oas_fragment):
    """Path response's example is not rendered if there's nothing to render."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                application/json:
                  schema:
                    type: object
                """
            ),
            "200",
        )
    )

    assert markup == ""


@pytest.mark.parametrize(
    ["status_code", "status_text"],
    [
        pytest.param("201", "Created", id="201"),
        pytest.param("307", "Temporary Redirect", id="307"),
        pytest.param("422", "Unprocessable Entity", id="422"),
    ],
)
def test_render_response_status_code(
    testrenderer, oas_fragment, status_code, status_text
):
    """Path response's example is rendered with proper status code."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                """
            ),
            status_code,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           HTTP/1.1 {status_code} {status_text}
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["status_range", "status_code", "status_text"],
    [
        pytest.param("2XX", "200", "OK", id="2XX"),
        pytest.param("3XX", "300", "Multiple Choices", id="3XX"),
        pytest.param("4XX", "400", "Bad Request", id="4XX"),
    ],
)
def test_render_response_status_code_range(
    testrenderer, oas_fragment, status_range, status_code, status_text
):
    """Path response's example is rendered with proper status range."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                """
            ),
            status_range,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           HTTP/1.1 {status_code} {status_text}
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


@pytest.mark.parametrize(
    ["status_code", "status_text"],
    [
        pytest.param("201", "Created", id="201"),
        pytest.param("307", "Temporary Redirect", id="307"),
        pytest.param("422", "Unprocessable Entity", id="422"),
    ],
)
def test_render_response_status_code_int(
    testrenderer, oas_fragment, status_code, status_text
):
    """Path response's example is rendered with proper status code."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                """
            ),
            status_code,
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           HTTP/1.1 {status_code} {status_text}
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )


def test_render_response_status_code_default(testrenderer, oas_fragment):
    """Path response's example is rendered when default is passed."""

    markup = textify(
        testrenderer.render_response_example(
            oas_fragment(
                """
                text/csv:
                  example: |
                    foo,baz
                    bar,42
                """
            ),
            "default",
        )
    )
    assert markup == textwrap.dedent(
        f"""\
        .. sourcecode:: http

           HTTP/1.1 000 Reason-Phrase
           Content-Type: text/csv

           foo,baz
           bar,42
        """.rstrip()
    )



"""OpenAPI spec renderer: render_parameter."""

import textwrap

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_render_parameter_path(testrenderer, oas_fragment):
    """Usual path parameter's definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: A unique evidence identifier to query.
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string, required
        """.rstrip()
    )


def test_render_parameter_path_minimal(testrenderer, oas_fragment):
    """Path parameter's minimal definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_description(testrenderer, oas_fragment):
    """Path parameter's 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: A unique evidence identifier to query.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_multiline_description(testrenderer, oas_fragment):
    """Path parameter's multiline 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: |
                  A unique evidence
                  identifier to query.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence
           identifier to query.
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_description_commonmark_default(
    testrenderer, oas_fragment
):
    """Path parameter's 'description' must be in commonmark by default."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence ``identifier``
           to **query**.
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_description_commonmark(fakestate, oas_fragment):
    """Path parameter's 'description' can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence ``identifier``
           to **query**.
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_description_restructuredtext(fakestate, oas_fragment):
    """Path parameter's 'description' can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence `identifier`
           to __query__.
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_deprecated(testrenderer, oas_fragment):
    """Path parameter's 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                deprecated: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: required, deprecated
        """.rstrip()
    )


def test_render_parameter_path_deprecated_false(testrenderer, oas_fragment):
    """Path parameter's 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                required: true
                deprecated: false
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_path_type(testrenderer, oas_fragment):
    """Path parameter's type is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_path_type_with_format(testrenderer, oas_fragment):
    """Path parameter's type with format is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                schema:
                  type: string
                  format: uuid4
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: string:uuid4
        """.rstrip()
    )


def test_render_parameter_path_type_from_content(testrenderer, oas_fragment):
    """Path parameter's type from content is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                content:
                  text/plain:
                    schema:
                      type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        :paramtype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_query(testrenderer, oas_fragment):
    """Usual query parameter's definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                description: A unique evidence identifier to query.
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
           A unique evidence identifier to query.
        :paramtype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_query_minimal(testrenderer, oas_fragment):
    """Query parameter's minimal definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: path
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :param evidenceId:
        """.rstrip()
    )


def test_render_parameter_query_description(testrenderer, oas_fragment):
    """Query parameter's 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                description: A unique evidence identifier to query.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
           A unique evidence identifier to query.
        """.rstrip()
    )


def test_render_parameter_query_multiline_description(testrenderer, oas_fragment):
    """Query parameter's multiline 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                description: |
                  A unique evidence
                  identifier to query.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
           A unique evidence
           identifier to query.
        """.rstrip()
    )


def test_render_parameter_query_description_commonmark_default(
    testrenderer, oas_fragment
):
    """Query parameter's 'description' must be in commonmark by default."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
           A unique evidence ``identifier``
           to **query**.
        """.rstrip()
    )


def test_render_parameter_query_description_commonmark(fakestate, oas_fragment):
    """Query parameter's 'description' can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
           A unique evidence ``identifier``
           to **query**.
        """.rstrip()
    )


def test_render_parameter_query_description_restructuredtext(fakestate, oas_fragment):
    """Query parameter's 'description' can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
           A unique evidence `identifier`
           to __query__.
        """.rstrip()
    )


def test_render_parameter_query_required(testrenderer, oas_fragment):
    """Query parameter's 'required' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                required: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: required
        """.rstrip()
    )


def test_render_parameter_query_required_false(testrenderer, oas_fragment):
    """Query parameter's 'required' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                required: false
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        """.rstrip()
    )


def test_render_parameter_query_deprecated(testrenderer, oas_fragment):
    """Query parameter's 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                deprecated: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: deprecated
        """.rstrip()
    )


def test_render_parameter_query_deprecated_false(testrenderer, oas_fragment):
    """Query parameter's 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                deprecated: false
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        """.rstrip()
    )


def test_render_parameter_query_required_deprecated(testrenderer, oas_fragment):
    """Both query parameter's markers are rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                required: true
                deprecated: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: required, deprecated
        """.rstrip()
    )


def test_render_parameter_query_type(testrenderer, oas_fragment):
    """Query parameter's type is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_query_type_with_format(testrenderer, oas_fragment):
    """Query parameter's type with format is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                schema:
                  type: string
                  format: uuid4
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: string:uuid4
        """.rstrip()
    )


def test_render_parameter_query_type_from_content(testrenderer, oas_fragment):
    """Query parameter's type from content is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: query
                content:
                  text/plain:
                    schema:
                      type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :queryparam evidenceId:
        :queryparamtype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_header(testrenderer, oas_fragment):
    """Usual header parameter's definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                description: A unique request identifier.
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
           A unique request identifier.
        :reqheadertype X-Request-Id: string
        """.rstrip()
    )


def test_render_parameter_header_minimal(testrenderer, oas_fragment):
    """Header parameter's minimal definition is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        """.rstrip()
    )


def test_render_parameter_header_description(testrenderer, oas_fragment):
    """Header parameter's 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                description: A unique request identifier.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
           A unique request identifier.
        """.rstrip()
    )


def test_render_parameter_header_multiline_description(testrenderer, oas_fragment):
    """Header parameter's multiline 'description' is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                description: |
                  A unique request
                  identifier.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
           A unique request
           identifier.
        """.rstrip()
    )


def test_render_parameter_header_description_commonmark_default(
    testrenderer, oas_fragment
):
    """Header parameter's 'description' must be in commonmark by default."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
           A unique evidence ``identifier``
           to **query**.
        """.rstrip()
    )


def test_render_parameter_header_description_commonmark(fakestate, oas_fragment):
    """Header parameter's 'description' can be in commonmark."""

    testrenderer = renderers.HttpdomainRenderer(fakestate, {"markup": "commonmark"})
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
           A unique evidence ``identifier``
           to **query**.
        """.rstrip()
    )


def test_render_parameter_header_description_restructuredtext(fakestate, oas_fragment):
    """Header parameter's 'description' can be in restructuredtext."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"markup": "restructuredtext"}
    )
    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                description: |
                  A unique evidence `identifier`
                  to __query__.
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
           A unique evidence `identifier`
           to __query__.
        """.rstrip()
    )


def test_render_parameter_header_required(testrenderer, oas_fragment):
    """Header parameter's 'required' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                required: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        :reqheadertype X-Request-Id: required
        """.rstrip()
    )


def test_render_parameter_header_required_false(testrenderer, oas_fragment):
    """Header parameter's 'required' marker is not rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                required: false
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        """.rstrip()
    )


def test_render_parameter_header_deprecated(testrenderer, oas_fragment):
    """Header parameter's 'deprecated' marker is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                deprecated: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        :reqheadertype X-Request-Id: deprecated
        """.rstrip()
    )


def test_render_parameter_header_deprecated_false(testrenderer, oas_fragment):
    """Header parameter's 'deprecated' marker is not rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                deprecated: false
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        """.rstrip()
    )


def test_render_parameter_header_required_deprecated(testrenderer, oas_fragment):
    """Both header parameter's markers are rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: X-Request-Id
                in: header
                required: true
                deprecated: true
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader X-Request-Id:
        :reqheadertype X-Request-Id: required, deprecated
        """.rstrip()
    )


def test_render_parameter_header_type(testrenderer, oas_fragment):
    """Header parameter's type is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                schema:
                  type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
        :reqheadertype evidenceId: string
        """.rstrip()
    )


def test_render_parameter_header_type_with_format(testrenderer, oas_fragment):
    """Header parameter's type with format is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                schema:
                  type: string
                  format: uuid4
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
        :reqheadertype evidenceId: string:uuid4
        """.rstrip()
    )


def test_render_parameter_header_type_from_content(testrenderer, oas_fragment):
    """Header parameter's type from content is rendered."""

    markup = textify(
        testrenderer.render_parameter(
            oas_fragment(
                """
                name: evidenceId
                in: header
                content:
                  text/plain:
                    schema:
                      type: string
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        :reqheader evidenceId:
        :reqheadertype evidenceId: string
        """.rstrip()
    )



"""OpenAPI spec renderer: render_restructuredtext_markup."""

import textwrap

from sphinxcontrib.openapi import renderers


def textify(generator):
    return "\n".join(generator)


def test_oas2_minimal(testrenderer, oas_fragment):
    """Minimal OAS 2 can be rendered."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                swagger: "2.0"
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      responses:
                        '200':
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           :statuscode 200:
              a response description
        """
    )


def test_oas2_complete(testrenderer, oas_fragment):
    """Feature rich OAS 2 can be rendered."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                swagger: "2.0"
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      responses:
                        '200':
                          description: a response description
                  /{username}:
                    parameters:
                      - in: path
                        name: username
                        required: true
                        type: string
                    get:
                      tags:
                        - tag_a
                        - tag_b
                      summary: an operation summary
                      description: an operation description
                      externalDocs: https://docs.example.com/
                      operationId: myOperation
                      produces:
                        - application/json
                      parameters:
                        - in: header
                          name: token
                          type: string
                        - in: query
                          name: id
                          type: string
                      responses:
                        '200':
                          schema:
                            items:
                              format: int32
                              type: integer
                            type: array
                          description: a response description
                        '404':
                          description: a username not found
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           :statuscode 200:
              a response description

        .. http:get:: /{username}

           **an operation summary**

           an operation description

           :reqheader token:
           :reqheadertype token: string
           :param username:
           :paramtype username: string, required
           :queryparam id:
           :queryparamtype id: string

           :statuscode 200:
              a response description

           :statuscode 404:
              a username not found
        """
    )


def test_oas2_schema_example(testrenderer, oas_fragment):
    """Schema's 'example' property can be used in example snippets."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                swagger: "2.0"
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      produces:
                        - application/json
                      responses:
                        '200':
                          schema:
                            example: |
                              [
                                19,
                                84
                              ]
                            items:
                              format: int32
                              type: integer
                            type: array
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description


           :statuscode 200:
              a response description

              .. sourcecode:: http

                 HTTP/1.1 200 OK
                 Content-Type: application/json

                 [
                   19,
                   84
                 ]
        """
    )


def test_oas2_complete_generate_examples_from_schema(fakestate, oas_fragment):
    """Schema can be used to generate example snippets."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"generate-examples-from-schemas": True}
    )
    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                swagger: "2.0"
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      produces:
                        - application/json
                      responses:
                        '200':
                          schema:
                            items:
                              format: int32
                              type: integer
                            type: array
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description


           :statuscode 200:
              a response description

              .. sourcecode:: http

                 HTTP/1.1 200 OK
                 Content-Type: application/json

                 [
                   1,
                   1
                 ]
        """
    )


def test_oas3_minimal(testrenderer, oas_fragment):
    """Minimal OAS 3 can be rendered."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      responses:
                        '200':
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           :statuscode 200:
              a response description
        """
    )


def test_oas3_complete(testrenderer, oas_fragment):
    """Feature rich OAS 3 can be rendered."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      responses:
                        '200':
                          description: a response description
                  /{username}:
                    parameters:
                      - in: path
                        name: username
                        schema:
                          type: string
                        required: true
                    get:
                      tags:
                        - tag_a
                        - tag_b
                      summary: an operation summary
                      description: an operation description
                      externalDocs: https://docs.example.com/
                      operationId: myOperation
                      parameters:
                        - in: header
                          name: token
                          schema:
                            type: string
                        - in: query
                          name: id
                          schema:
                            type: string
                      responses:
                        '200':
                          content:
                            application/json:
                              schema:
                                items:
                                  format: int32
                                  type: integer
                                type: array
                          description: a response description
                        '404':
                          description: a username not found
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           :statuscode 200:
              a response description

        .. http:get:: /{username}

           **an operation summary**

           an operation description

           :reqheader token:
           :reqheadertype token: string
           :param username:
           :paramtype username: string, required
           :queryparam id:
           :queryparamtype id: string

           :statuscode 200:
              a response description

           :statuscode 404:
              a username not found
        """
    )


def test_oas3_schema_example(testrenderer, oas_fragment):
    """Schema's 'example' property can be used in example snippets."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      responses:
                        '200':
                          content:
                            application/json:
                              schema:
                                example: |
                                  [
                                    19,
                                    84
                                  ]
                                items:
                                  format: int32
                                  type: integer
                                type: array
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description


           :statuscode 200:
              a response description

              .. sourcecode:: http

                 HTTP/1.1 200 OK
                 Content-Type: application/json

                 [
                   19,
                   84
                 ]
        """
    )


def test_oas3_generate_examples_from_schema(fakestate, oas_fragment):
    """Schema can be used to generate example snippets."""

    testrenderer = renderers.HttpdomainRenderer(
        fakestate, {"generate-examples-from-schemas": True}
    )
    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      responses:
                        '200':
                          content:
                            application/json:
                              schema:
                                items:
                                  format: int32
                                  type: integer
                                type: array
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description


           :statuscode 200:
              a response description

              .. sourcecode:: http

                 HTTP/1.1 200 OK
                 Content-Type: application/json

                 [
                   1,
                   1
                 ]
        """
    )


def test_oas3_request_body(testrenderer, oas_fragment):
    """Request body example is rendered."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      requestBody:
                        content:
                          application/json:
                            examples:
                              test:
                                value:
                                  foo: bar
                                  baz: 42
                      responses:
                        '200':
                          description: a response description
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description

           .. sourcecode:: http

              GET /test HTTP/1.1
              Content-Type: application/json

              {
                "foo": "bar",
                "baz": 42
              }

           :statuscode 200:
              a response description
        """
    )


def test_oas3_response_example_2xx(testrenderer, oas_fragment):
    """Response examples are rendered for 2XX status codes."""

    markup = textify(
        testrenderer.render_restructuredtext_markup(
            oas_fragment(
                """
                openapi: 3.0.3
                info:
                  title: An example spec
                  version: 1.0
                paths:
                  /test:
                    get:
                      description: an operation description
                      responses:
                        '200':
                          content:
                            application/json:
                              example: |
                                [
                                  19,
                                  84
                                ]
                          description: a response description
                        '404':
                          content:
                            application/json:
                              example: |
                                {
                                  "message": "an error message"
                                }
                          description: resource not found
                """
            )
        )
    )
    assert markup == textwrap.dedent(
        """\
        .. http:get:: /test

           an operation description

           :statuscode 200:
              a response description

              .. sourcecode:: http

                 HTTP/1.1 200 OK
                 Content-Type: application/json

                 [
                   19,
                   84
                 ]
           :statuscode 404:
              resource not found
        """
    )



import argparse
import logging

from sphinxcontrib.openapi import directive, renderers


def main():
    parser = argparse.ArgumentParser(
        prog='oas2rst',
        description='Export OpenAPI Specification files to reStructuredText \
            files')
    parser.add_argument(
        "-l", "--level",
        action='store',
        default=logging.INFO,
        dest='level',
        help="Logging level")
    parser.add_argument(
        "-e", "--encoding",
        action='store',
        default="UTF-8",
        dest='encoding',
        help="Source file encoding")
    parser.add_argument(
        "-p", "--paths",
        action='append',
        dest='paths',
        help="Endpoints to be rendered")
    parser.add_argument(
        "-x", "--examples",
        action='store_true',
        dest='examples',
        help="Include examples")
    parser.add_argument(
        "--group-examples",
        action='store_true',
        dest='group_examples',
        help="Group examples after the operation description")
    parser.add_argument(
        "-E", "--entities",
        action='store_true',
        dest='entities',
        help="Include links to entities")
    parser.add_argument(
        "-g", "--group",
        action='store_true',
        dest='group',
        help="Group paths by tag")
    parser.add_argument(
        "-i", "--input",
        dest='input',
        required=True,
        help="Input file")
    parser.add_argument(
        "-o", "--output",
        type=argparse.FileType('w'),
        required=True,
        dest='output',
        help="Output file")

    options = parser.parse_args()
    logging.getLogger().setLevel(options.level)

    openapi_options = {}
    if options.paths:
        openapi_options['paths'] = options.paths
    if options.examples:
        openapi_options['examples'] = True
    if options.group_examples:
        openapi_options['group_examples'] = True
    if options.entities:
        openapi_options['entities'] = True
    if options.group:
        openapi_options['group'] = True

    openapi_options.setdefault('uri', 'file://%s' % options.input)
    spec = directive._get_spec(options.input, options.encoding)
    renderer = renderers.HttpdomainOldRenderer(None, openapi_options)

    for line in renderer.render_restructuredtext_markup(spec):
        options.output.write(line+'\n')
        logging.debug(line)


if __name__ == '__main__':
    main()



"""OpenAPI schema utility functions."""

from io import StringIO


_DEFAULT_EXAMPLES = {
    "string": "string",
    "integer": 1,
    "number": 1.0,
    "boolean": True,
    "array": [],
}


_DEFAULT_STRING_EXAMPLES = {
    "date": "2020-01-01",
    "date-time": "2020-01-01T01:01:01Z",
    "password": "********",
    "byte": "QG1pY2hhZWxncmFoYW1ldmFucw==",
    "ipv4": "127.0.0.1",
    "ipv6": "::1",
}


def example_from_schema(schema):
    """
    Generates an example request/response body from the provided schema.

    >>> schema = {
    ...     "type": "object",
    ...     "required": ["id", "name"],
    ...     "properties": {
    ...         "id": {
    ...             "type": "integer",
    ...             "format": "int64"
    ...         },
    ...         "name": {
    ...             "type": "string",
    ...             "example": "John Smith"
    ...         },
    ...         "tag": {
    ...             "type": "string"
    ...         }
    ...     }
    ... }
    >>> example = example_from_schema(schema)
    >>> assert example == {
    ...     "id": 1,
    ...     "name": "John Smith",
    ...     "tag": "string"
    ... }
    """
    # If an example was provided then we use that
    if "example" in schema:
        return schema["example"]

    elif "oneOf" in schema:
        return example_from_schema(schema["oneOf"][0])

    elif "anyOf" in schema:
        return example_from_schema(schema["anyOf"][0])

    elif "allOf" in schema:
        # Combine schema examples
        example = {}
        for sub_schema in schema["allOf"]:
            example.update(example_from_schema(sub_schema))
        return example

    elif "enum" in schema:
        return schema["enum"][0]

    elif "type" not in schema:
        # Any type
        return _DEFAULT_EXAMPLES["integer"]

    elif schema["type"] == "object" or "properties" in schema:
        example = {}
        for prop, prop_schema in schema.get("properties", {}).items():
            example[prop] = example_from_schema(prop_schema)
        return example

    elif schema["type"] == "array":
        items = schema["items"]
        min_length = schema.get("minItems", 0)
        max_length = schema.get("maxItems", max(min_length, 2))
        assert min_length <= max_length
        # Try generate at least 2 example array items
        gen_length = min(2, max_length) if min_length <= 2 else min_length

        example_items = []
        if items == {}:
            # Any-type arrays
            example_items.extend(_DEFAULT_EXAMPLES.values())
        elif isinstance(items, dict) and "oneOf" in items:
            # Mixed-type arrays
            example_items.append(_DEFAULT_EXAMPLES[sorted(items["oneOf"])[0]])
        else:
            example_items.append(example_from_schema(items))

        # Generate array containing example_items and satisfying min_length and max_length
        return [example_items[i % len(example_items)] for i in range(gen_length)]

    elif schema["type"] == "string":
        example_string = _DEFAULT_STRING_EXAMPLES.get(
            schema.get("format", None), _DEFAULT_EXAMPLES["string"]
        )
        min_length = schema.get("minLength", 0)
        max_length = schema.get("maxLength", max(min_length, len(example_string)))
        gen_length = (
            min(len(example_string), max_length)
            if min_length <= len(example_string)
            else min_length
        )
        assert 0 <= min_length <= max_length
        if min_length <= len(example_string) <= max_length:
            return example_string
        else:
            example_builder = StringIO()
            for i in range(gen_length):
                example_builder.write(example_string[i % len(example_string)])
            example_builder.seek(0)
            return example_builder.read()

    elif schema["type"] in ("integer", "number"):
        example = _DEFAULT_EXAMPLES[schema["type"]]
        if "minimum" in schema and "maximum" in schema:
            # Take average
            example = schema["minimum"] + (schema["maximum"] - schema["minimum"]) / 2
        elif "minimum" in schema and example <= schema["minimum"]:
            example = schema["minimum"] + 1
        elif "maximum" in schema and example >= schema["maximum"]:
            example = schema["maximum"] - 1
        return float(example) if schema["type"] == "number" else int(example)

    else:
        return _DEFAULT_EXAMPLES[schema["type"]]
